/* -*- Mode: C++; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 *   Copyright 2020-2021 Couchbase, Inc.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

#pragma once

#include "client_opcode.hxx"
#include "cmd_info.hxx"
#include "core/document_id.hxx"
#include "core/impl/subdoc/command.hxx"
#include "core/io/mcbp_message.hxx"
#include "status.hxx"

#include <gsl/assert>

namespace couchbase::core::protocol
{

class lookup_in_replica_response_body
{
  public:
    static const inline client_opcode opcode = client_opcode::subdoc_multi_lookup;

    struct lookup_in_field {
        key_value_status_code status{};
        std::string value;
    };

  private:
    std::vector<lookup_in_field> fields_{};

  public:
    [[nodiscard]] const std::vector<lookup_in_field>& fields() const
    {
        return fields_;
    }

    [[nodiscard]] bool parse(key_value_status_code status,
                             const header_buffer& header,
                             std::uint8_t framing_extras_size,
                             std::uint16_t key_size,
                             std::uint8_t extras_size,
                             const std::vector<std::byte>& body,
                             const cmd_info& info);
};

class lookup_in_replica_request_body
{
  public:
    using response_body_type = lookup_in_replica_response_body;
    static const inline client_opcode opcode = client_opcode::subdoc_multi_lookup;

    /**
     * Tells the server to operate on replica vbucket instead of active
     */
    static const inline std::uint8_t doc_flag_replica_read = 0b0010'0000;

  private:
    std::vector<std::byte> key_;
    std::vector<std::byte> extras_{};
    std::vector<std::byte> value_{};

    std::uint8_t flags_{ 0 };
    std::vector<couchbase::core::impl::subdoc::command> specs_;

  public:
    void id(const document_id& id);

    void read_replica(bool value)
    {
        if (value) {
            flags_ = flags_ | doc_flag_replica_read;
        }
    }

    void specs(const std::vector<couchbase::core::impl::subdoc::command>& specs)
    {
        specs_ = specs;
    }

    [[nodiscard]] const auto& key() const
    {
        return key_;
    }

    [[nodiscard]] const auto& framing_extras() const
    {
        return empty_buffer;
    }

    [[nodiscard]] const auto& extras()
    {
        if (extras_.empty()) {
            fill_extras();
        }
        return extras_;
    }

    [[nodiscard]] const auto& value()
    {
        if (value_.empty()) {
            fill_value();
        }
        return value_;
    }

    [[nodiscard]] std::size_t size()
    {
        if (extras_.empty()) {
            fill_extras();
        }
        if (value_.empty()) {
            fill_value();
        }
        return key_.size() + extras_.size() + value_.size();
    }

  private:
    void fill_extras();

    void fill_value();
};

} // namespace couchbase::core::protocol
